/* Class to maintain a grey-scale 2D image and write it to disk.

   Code adapted from Paul Haeberli's tutorial:
	http://reality.sgi.com/grafica/sgiimage.html
*/

#include "ImageSGI.h"    

ImageSGI::ImageSGI(int w, int h, double x1, double y1, double x2, double y2)
{
	height = h;
	width = w;

	min_x = x1;
	min_y = y1;
	max_x = x2;
	max_y = y2;

	pixels = new unsigned char *[width];

	for (int i=0; i<width; i++)
		pixels[i] = new unsigned char[height];
	reset();
}

ImageSGI::ImageSGI(int w, int h)
{
	height = h;
	width = w;

	min_x = 0.0;
	min_y = 0.0;
	max_x = 1.0;
	max_y = 1.0;

	pixels = new unsigned char *[width];

	for (int i=0; i<width; i++)
		pixels[i] = new unsigned char[height];

	reset();
}
 
void ImageSGI::putbyte(FILE *outf, unsigned char val)
    {
	unsigned char buf[1];
     
	buf[0] = val;
	fwrite(buf,1,1,outf);
    }
     
void ImageSGI::putshort(FILE *outf, unsigned short val)
    {
	unsigned char buf[2];
     
	buf[0] = (val>>8);
	buf[1] = (val>>0);
	fwrite(buf,2,1,outf);
    }
     
int ImageSGI::putlong(FILE *outf, unsigned long val)
    {
	unsigned char buf[4];
     
	buf[0] = (val>>24);
	buf[1] = (val>>16);
	buf[2] = (val>>8);
	buf[3] = (val>>0);
	return fwrite(buf,4,1,outf);
    }
     
void ImageSGI::write(char *filename)
    {
	FILE *of;
	char iname[80];
	unsigned char outbuf[width];
	int i, x, y;
     
	of = fopen(filename,"w");
	if(!of) {
	    fprintf(stderr,"sgiimage: can't open output file\n");
	    exit(1);
	}
	putshort(of,474);       /* MAGIC               */
	putbyte(of,0);          /* STORAGE is VERBATIM */
	putbyte(of,1);          /* BPC is 1            */
	putshort(of,2);         /* DIMENSION is 2      */
	putshort(of,width);    /* XSIZE               */
	putshort(of,height);    /* YSIZE               */
	putshort(of,1);         /* ZSIZE               */
	putlong(of,0);          /* PIXmin is 0         */
	putlong(of,255);        /* PIXMAX is 255       */
	for(i=0; i<4; i++)      /* DUMMY 4 bytes       */
	    putbyte(of,0);
	strcpy(iname,"No Name");
	fwrite(iname,80,1,of);  /* IMAGENAME           */
	putlong(of,0);          /* COLORMAP is 0       */
	for(i=0; i<404; i++)    /* DUMMY 404 bytes     */
	    putbyte(of,0);
     
	for (y = 0; y<height; y++)
	{
		for (x = 0; x<width; x++)
		{
			outbuf[x] = pixels[x][y];

		}
		fwrite(outbuf,width,1,of);
	}

	fclose(of);
    }

// set all the pixels to zero
void ImageSGI::reset()
{
	for (int x=0; x<width; x++)
		for (int y=0; y<height; y++)
			pixels[x][y] = 0;
}

// draw a vertical line in our pixel array
// we assume y1 < y2
void ImageSGI::verticalLine(double x, double y1, double y2, int color)
{
	int i, jStart, jEnd;

	i = hashI(x);

	jStart = hashJ(y1);
	jEnd = hashJ(y2);

	for (int k = jStart; k <= jEnd; k++)
	{
		// check to make sure the computed indices are in range
		if (!((i<0) || (i>=width) || (k<0) || (k>=height)))
			pixels[i][k] = color;
	}
}

// draw a horizontal line in our pixel array
// we assume y1 < y2
void ImageSGI::horizontalLine(double y, double x1, double x2, int color)
{
	int j, iStart, iEnd;

	j = hashJ(y);
	
	iStart = hashI(x1);
	iEnd = hashI(x2);

	for (int k = iStart; k <= iEnd; k++)
	{
		// check to make sure the computed indices are in range
		if (!((k<0) || (k>=width) || (j<0) || (j>=height)))
			pixels[k][j] = color;
	}
}


// increase the pixel color of the given point on the grid
void ImageSGI::add(double x, double y)
{
	int i, j;

	i = hashI(x);
	j = hashJ(y);

	// check to make sure the computed indices are in range
	if (!((i<0) || (i>=width) || (j<0) || (j>=height)))
		pixels[i][j]++;
}

// increase the pixel color of the given point on the grid
void ImageSGI::add(int x, int y)
{
	// check to make sure the computed indices are in range
	if (!((x<0) || (x>=width) || (y<0) || (y>=height)))
		pixels[x][y]++;	
}

void ImageSGI::add(double x, double y, int color)
{
	int i, j;

	i = hashI(x);
	j = hashJ(y);

	// check to make sure the computed indices are in range
	if (!((i<0) || (i>=width) || (j<0) || (j>=height)))
		pixels[i][j] = color;
}	

int ImageSGI::hashI(double x)
{
	return (int) (width * (x-min_x) / (max_x - min_x));
}

int ImageSGI::hashJ(double y)
{
	return (int) (height * (y-min_y) / (max_y - min_y));

}

void ImageSGI::add(int x, int y, int color)
{
	// check to make sure the computed indices are in range
	if (!((x<0) || (x>=width) || (y<0) || (y>=height)))
		pixels[x][y] = (unsigned char) color;
}	

// scale the color values so we see them better
void ImageSGI::scale(int min)
{
	int range = 255-min;

	int max = 0;

	for (int i=0; i<width; i++)
		for (int j=0; j<height; j++)
		{
			if (pixels[i][j] > max)
				max = pixels[i][j];
		}

	double mult = (double) range / (double) max;

	for (int i=0; i<width; i++)
		for (int j=0; j<height; j++)
		{
			pixels[i][j] = (char) (pixels[i][j]*mult + min);
		}
}
