#include "Master.h"
#include "msgtags.h"

Master::Master(int num_p, int port, int dataport, int debug)
{
	if (debug>1)
		DEBUG = 1;
	else
		DEBUG = 0;

	world = new World();
	mpi_link = new MasterMPI(num_p, debug, world);
	client_link = new ClientLink(num_p, port, dataport, debug, world, mpi_link);
}


void Master::connectInit()
{
	client_link->connectInit();
}

void Master::loop()
{
	int bye = 0;
	int command;
	int connected = 0;
	Timer watch;

	// You can use the following section of code to initialize a simulation
	// and run it without using the client application.

	if (CLIENT_MODE)
	{

		// avoid doing the client side tango for testing purposes
		connected=1;		

		// set the parameters to some default values

		world->param.numPts 		= 4000;
		world->param.velocityScale   	= 5.0E-5;
		world->param.initVel.setXY(0.0, 0.0);		
		world->param.machNum      	= 0.1;		
		world->param.initDensity	= 1000.0;	
		world->param.nu 		= 1.0e-6;
		world->param.deltaT		= 2.0E-3;
		world->param.numSteps		= 500000;
		world->param.outputEvery	= 5000;
		world->param.firstOutput	= 0;
		world->param.densityVariation   = 0.03;
		world->param.lengthScale	= 1.0E-3;
		world->param.bodyFX 		= 0.05E-5;
		world->param.bodyFY		= 0.0;
			
		world->setContainerHeight(0.15);
		world->setContainerWidth(0.15);

		// Now add out shapes to the grid.
		//But first tell the grid how many to expect.
		world->theGrid->setNumShapes(3);
		double attributes[MAX_ATTRS];

		attributes[0] = CYLINDER;
		attributes[1] = 0.075;
		attributes[2] = 0.075;
		attributes[3] = 0.02;
		 
		world->theGrid->addShape(attributes);

		attributes[0] = BOX;
		attributes[1] = 0;
		attributes[2] = 0;
		attributes[3] = 0.15;
		attributes[4] = 0.025;

		world->theGrid->addShape(attributes);

		attributes[0] = BOX;
		attributes[1] = 0;
		attributes[2] = 0.125;
		attributes[3] = 0.15;
		attributes[4] = 0.025;
		world->theGrid->addShape(attributes);
	
		cout << "Starting in stand alone mode (no client)..." << endl;

	}

	while (!bye)
	{
		// if we are not connected wait for a connection
		if (!connected)
		{
			connectInit();
			connected = 1;
		}

		if (DEBUG) 
			printf("Master waiting for command...\n");

		if (CLIENT_MODE)
			command = RUNSIM;
		else
			command = client_link->getCommand();

		if (command == BYE)
		{
			// tell the slaves this sim is over
			mpi_link->sendCommand(command);
			client_link->bye();
			connected = 0;
			if (DEBUG) 
				printf("Master was told goodbye.\n");
		}
		else if (command == SHUTDOWN)
		{
			// tell the slave to finalize and shutdown
			client_link->bye();

			// tell the slaves this sim is over
			mpi_link->sendCommand(command);
			connected = 0;
			bye = 1;		
			if (DEBUG) 
				printf("Master was told to shutdown.\n");
		}
		else if (command == STATS)
		{
			// run stats on everybody

			double results[mpi_link->num_p];
			char hostname[MAX_NODE][80];
			
			// blank out that array
			for (int j=0; j<MAX_NODE; j++)
				for (int i=0; i<80; i++)
					hostname[j][i] = '\0';

			mpi_link->runStats(results, hostname);
			client_link->sendStats(results, hostname);

		}
		else if (command == RUNSIM)
		{
			watch.startTimer();

			// test out running simulation on just master

			// get an update of params, world and network
			// before we start this thing

			if (!CLIENT_MODE)
				client_link->recvAll();

			world->createWorld();	// init our grid

			// now we have to get the slaves ready to rock

			if (DEBUG)
				cout << "Initing the slaves..." << endl;

			mpi_link->sendCommand(command);
			mpi_link->initSlaves();

			// loop until the slaves send us -1
			// changed, kdv 5/10
			int i= 0;
			while (mpi_link->recvUpdate() > 0)
			{
				if (DEBUG)
					cout << "Recv'd update #" << i << endl;
				
				if (OUTPUT_DEBUG)
					world->writeFileDebug();
				if (OUTPUT_IMAGE)
					world->writeFileImage();
                                if (OUTPUT_CLIENT)
                                        world->writeFileClient();
				if (OUTPUT_NEIGHBOR)
					world->writeFileNeighbor();

				world->incrementTime();

				// now send it off to the clinet

				if (!CLIENT_MODE)		
					client_link->sendClientFrame();				
				
				// print out the biggest difference in our particle densities
				world->theGrid->printMinMaxDensity();

				i++;
			}

			if (!CLIENT_MODE)
				client_link->simDone();
			else
			{
				// tell the slaves this sim is over
				mpi_link->sendCommand(SHUTDOWN);

				connected = 0;
				bye = 1;		
			}
	
			watch.stopTimer();

			cout << endl;
			cout << "Total elapsed time = " << watch.elapsed << endl;
			cout << "num_p = " << mpi_link->num_p << endl;
	
		        ofstream fout("data.txt", ios::ate | ios::out);
		        fout << watch.elapsed << "   ";
			fout.close();

			// destroy everything and get ready for a new run
                        delete world;
                        world = new World();
                        mpi_link->world = world;
                        client_link->world = world;
		}
	}

}
