#include "MasterMPI.h"
#include "mpi.h"
#include <stdio.h>
#include "msgtags.h"
#include <unistd.h>		// for gethostname
#include "Bench.h"

MasterMPI::MasterMPI(int p, int debug, World *w)
{
	if (debug>2)
		DEBUG = 1;
	else
		DEBUG = 0;

	num_p = p;
	num_p_compute = p;

	world = w;

	updateTag = UPDATE_OFFSET;
}

// utility function, sends the given command to all the nodes

void MasterMPI::sendCommand(int command)
{
	int A[1];

	A[0] = command;

	if (DEBUG)
		printf("Master sending command=%d to slaves...\n", command);

	for (int i=1; i<num_p; i++)
		MPI_Send(A, 1, MPI_INT, i, COMMAND, MPI_COMM_WORLD);
	
}


// run a performance benchmark on all our nodes
// put the results in the result array, put the machine
// names in the 2-d char array hostname

void MasterMPI::runStats(double result[], char hostname[MAX_NODE][80])
{
	if (DEBUG)
		printf("Telling everyone to run benchmark\n");

	int i;
	MPI_Status status;

	sendCommand(STATS);

	if (DEBUG)
		printf("Running benchmark\n");

	double flops = 0.0;
	Bench bench;

	flops = bench.measureFlops();

	/* master collects results from slaves */

        result[0] = (double) flops;

        gethostname(hostname[0], 80);

        if (DEBUG)
        	printf("Gathering results...\n");

        for (int source=1; source<num_p; source++)
        {
         	MPI_Recv(&result[source],1,MPI_DOUBLE,source,TAG,
                                MPI_COMM_WORLD, &status);

                MPI_Recv(hostname[source],80,MPI_CHAR,source,TAG,
                                MPI_COMM_WORLD, &status);
        }

	for (i=0; i<num_p; i++)
		hostname[i][strlen(hostname[i])] = '\n';
}




// tell each slave what it needs to know to setup their SlaveWorld's

void MasterMPI::initSlaves()
{
	// tell the slaves we are going to run sim, so it is
	// time to initialize them

	int i, j, k;
	
	int A[1];

	// next lets get everybody a copy of the sim parameters

	double D[PARAM_SIZE];
	int max_particles;

	// allocate a buffer array we will use to send all the particles to the slave

	if (world->theGrid->getNumPts() > world->theGrid->getNumShapePts())
		max_particles = world->theGrid->getNumPts();
	else
		max_particles = world->theGrid->getNumShapePts();

	double particles[SMOOTH_SIZE * max_particles];

	world->param.getParams(D);

	if (DEBUG)
		cout << "Sending sim parameters to all the slaves..." 
		     << endl;

	for (i=1; i<num_p; i++)
	{
		MPI_Send(D, PARAM_SIZE, MPI_DOUBLE, i, TAG, MPI_COMM_WORLD);
	}
	
	int num_rows = world->theGrid->getNumCellsHigh();

	// If we have more slaves than we have rows in the grid, then
	// we need to tell some of them to chill out.  

	if (num_p > num_rows)
	{
		num_p_compute = num_rows;

		// we will tell the upper slaves they are not needed
		A[0] = 0;
		
		for (int k = num_rows; k<num_p; k++) 		
			MPI_Send(A, 1, MPI_INT, k, TAG, MPI_COMM_WORLD);		
	}

	if (DEBUG) cout << "We have " << num_p-1
	                << " slaves, we are using " << num_p_compute-1 
					<< " for the simulation." << endl;

	// figure out who gets what rows of our grid

	int each, extra;
	
	// prevent div by 0 if only 1 processor

	if (num_p!=1)
	{	
	 	each = num_rows / (num_p_compute - 1);
		extra = num_rows % (num_p_compute - 1);
	}

	double initConds[9];
	int start, end, pos;

	int current = 0;
	for (i=1; i<num_p_compute; i++)
	{
		// first tell this processor that they are indeed needed
		A[0] = 1;
		MPI_Send(A, 1, MPI_INT, i, TAG, MPI_COMM_WORLD);		

		// also tell them how many total slaves are computing
		A[0] = num_p_compute;
		MPI_Send(A, 1, MPI_INT, i, TAG, MPI_COMM_WORLD);		

		start = current;
		current += each;
		if (i > (num_p_compute - 1) - extra)
			current++;
		
		// this if avoid an extra row on top being created
		if (i == num_p_compute - 1)
			end = current - 1;
		else
			end = current + 1;

		if (DEBUG)
			cout << "Telling slave = " << i << " start = " 
			     << start << " end = " << end << endl;

		// now send this to this particular slave
		initConds[0] = start;
		initConds[1] = end;
		initConds[2] = world->theGrid->getNumPts();
		initConds[3] = world->theGrid->getNumShapePts();
		initConds[4] = world->theGrid->getNumCellsWide();
		initConds[5] = world->theGrid->getNumCellsHigh();
		initConds[6] = world->theGrid->getCellSize();
		initConds[7] = world->theGrid->getContainerWidth();
		initConds[8] = world->theGrid->getContainerHeight();
		
		MPI_Send(initConds, 9, MPI_DOUBLE, i, TAG, MPI_COMM_WORLD);

		// now we must send all of the points that belong to this processor
		// since the slaves have the parameters, we can get by with
		// just sending the x, y position of the points

		pos = 0;
		SmoothParticle sp;
		int hashJ;

		for (j=0; j<world->theGrid->getNumPts(); j++)
		{
			sp = world->theGrid->getParticle(j);

			// check to see if particle hashes into our reponsibility

			hashJ = world->theGrid->hashJ(sp.getPositionY());

			if ((hashJ>=start) && (hashJ<=end) && (!sp.isEmpty()))
			{
				particles[pos] = sp.getPositionX();
				pos++;
				particles[pos] = sp.getPositionY();
				pos++;
				particles[pos] = (double) sp.ID;
				pos++;
			}	
		}

		// tell the slave how many doubles to expect

		A[0] = pos;
		MPI_Send(A, 1, MPI_INT, i, TAG, MPI_COMM_WORLD);

		// now send the big buffer to the slave

		if (DEBUG)
			cout << "Master sending " << pos 
			     << " doubles to slave (free particles)..." << endl;

		MPI_Send(particles, pos, MPI_DOUBLE, i, TAG, MPI_COMM_WORLD);

		// send the boundary particles that belong to each processor

		pos = 0;

		for (j=0; j<world->theGrid->getNumShapePts(); j++)
		{
			sp = world->theGrid->getShapeParticle(j);

			// check to see if particle hashes into our reponsibility

			hashJ = world->theGrid->hashJ(sp.getPositionY());
			if ((hashJ>=start) && (hashJ<=end) && (!sp.isEmpty()))
			{
				particles[pos] = sp.getPositionX();
				pos++;
				particles[pos] = sp.getPositionY();
				pos++;
				particles[pos] = sp.ID;
				pos++;
				particles[pos] = sp.getShapeID();
				pos++;
			}	
		}

		// tell the slave how many doubles to expect

		A[0] = pos;
		MPI_Send(A, 1, MPI_INT, i, TAG, MPI_COMM_WORLD);

		// now send the big buffer to the slave

		if (DEBUG)
			cout << "Master sending " << pos << " doubles to slave (boundary points)..." << endl;

		MPI_Send(particles, pos, MPI_DOUBLE, i, TAG, MPI_COMM_WORLD);

		// We need to send the slaves the shapes themselves.
		// Tell the slaves how many shapes to expect.
		A[0] = world->theGrid->getNumShapes();
		MPI_Send(A, 1, MPI_INT, i, TAG, MPI_COMM_WORLD);
		
		// now allocate enough room in an array for all the
		// attributes of all of the shapes.
		double * all_attributes;
		all_attributes = new double[MAX_ATTRS*A[0]];
  	
  		//Now loop over the shapes and collect their attributes
		// in 'all_attribute'.
		for(j = 0; j < A[0]; j++)
			world->theGrid->getShapeAttributes(j, 
								all_attributes+j*MAX_ATTRS);
  		
		// The slaves can figure out how many attributes are coming
		// so let's just send `em.
		MPI_Send(all_attributes, MAX_ATTRS*A[0], MPI_DOUBLE, i, 
		                                       TAG, MPI_COMM_WORLD);

		int B[1];  
		// send the slave a value to start its globalID field at
		B[0] = world->globalID;
		MPI_Send(B, 1, MPI_INT, i, TAG, MPI_COMM_WORLD);
	}
}

// receive a grid update from all the slaves

int MasterMPI::recvUpdate()
{
	MPI_Status status;
	int num_pts, size;

	SmoothParticle sp;
	int pos, i, j, k;

	// clear out our grid
	world->theGrid->deleteAllFree();

        for (k=1; k<num_p_compute; k++)
        {
		// find out how many pts this slave is sending us

		int error;
		int stupid[2];       // don't ask me why, but doing 
				     // &num_pts breaks MPI_Recv
		int source;

		// receive from whoever gets here first, we use a different
		// tag number for each master/slave update.  
 
         	error = MPI_Recv(stupid,1,MPI_INT,MPI_ANY_SOURCE, updateTag,
                              MPI_COMM_WORLD, &status);

		source = status.MPI_SOURCE;

		if (error!=MPI_SUCCESS)
			cout << "MPI_Recv failed!" << endl;

		num_pts = stupid[0];

		// receive all the body particles from this slave

		if (num_pts>0)
		{
			size = num_pts * SMOOTH_SIZE;

			double buffer[size];

               		MPI_Recv(buffer,size,MPI_DOUBLE,source,PTS_UPDATE,
                                MPI_COMM_WORLD, &status);

			if (DEBUG)
				cout << "Recv'd " << num_pts 
			          	<< " body particles from " << source << endl;
			
			// now we better add these babies to our grid

			pos = 0;
	            	for (i=0; i<num_pts; i++)
        	    	{
    		                // now create a new particle and add to grid

	           		sp.setPacket(&buffer[i*SMOOTH_SIZE]);


				// add it to the grid, we will just use the same 
				// particle for plus and minus  one half since we 
				// don't really care about these on the master
				world->theGrid->addNewMigrantParticle(sp, sp, sp);  
		      
			} // end for loop
		} // end if conditional
        } // end for loop

	updateTag++;

	// We return 0 if the slaves have given us -1 number of points
	// This indicates that the slaves want to shutdown, this will
	// make sure we shutdown correctly.

	if (num_pts == -1)
		return 0;
	else
		return 1;
}
