#include "SmoothParticle.h"
#include <iostream.h>

//******************************************************************
// Default constructor.
//******************************************************************
SmoothParticle::SmoothParticle() : Particle(), pressure(0), density(0),
h(0), c(0), ChangeDensity(0), ChangeVelocity(0, 0), ID(0), shapeID(-1),
minDensity(999999.0), maxDensity(-999999.0), numNeighbors(0)
{}


//******************************************************************
// Constructor.
//******************************************************************
SmoothParticle::SmoothParticle(Point p, Point v, double m, double d, 
                              double pr, double len, double speedSnd) : 
Particle(p, v, m), pressure(pr), density(d), h(len), c(speedSnd), 
ChangeDensity(0.0), ChangeVelocity(0.0,0.0), shapeID(-1), ID(0),
minDensity(999999.0), maxDensity(-999999.0), numNeighbors(0)
{}



//******************************************************************
// Copy constructor
//******************************************************************
SmoothParticle::SmoothParticle(const SmoothParticle & SP)
{
	pos = SP.getPosition(); 
	vel = SP.getVelocity();
	mass = SP.getMass(); 
	density = SP.getDensity();
	pressure = SP.getPressure();
	h = SP.getSmoothingLength();
	c = SP.getSpeedSound();

	minDensity = SP.getMinDensity();
	maxDensity = SP.getMaxDensity();

	ChangeVelocity = SP.getChangeVelocity();
	ChangeDensity = SP.getChangeDensity();
	deleted = SP.deleted;
	ID = SP.ID;
	shapeID = SP.shapeID;
	numNeighbors = SP.numNeighbors;
}


//******************************************************************
// overloaded "=" operator
//******************************************************************
SmoothParticle & SmoothParticle::operator = (const SmoothParticle & SP)
{
	// A check to make sure we're not trying to copy ourself to
	// ourself. If this did happen, then we would delete all
	// information before we had a chance to copy it.
	if(this == &SP)
		return *this;

	pos = SP.getPosition(); 
	vel = SP.getVelocity();
	mass = SP.getMass(); 
	density = SP.getDensity();
	pressure = SP.getPressure();
	h = SP.getSmoothingLength();
	c = SP.getSpeedSound();

	minDensity = SP.getMinDensity();
	maxDensity = SP.getMaxDensity();

	ChangeVelocity = SP.getChangeVelocity();
	ChangeDensity = SP.getChangeDensity();
	deleted = SP.deleted;		
	ID = SP.ID;
	shapeID = SP.shapeID;
	numNeighbors = SP.numNeighbors;
	
	return (*this);
}


//******************************************************************
//  overloaded ostream operator (standard output)
//******************************************************************
ostream & operator << ( ostream & str, const SmoothParticle & sp)
{
	str  << " pos = " << sp.pos;

	// extend for debugging purposes
	str  << " vel = " << sp.vel;
	str  << " mass = " << sp.mass;
	str  << " pressure = " << sp.pressure;
	str  << " density = " << sp.density;
	str  << " h = " << sp.h;
	str  << " c = " << sp.c;
	str  << " minDensity = " << sp.minDensity;
	str  << " maxDensity = " << sp.maxDensity;
	str  << " ChangeDensity " << sp.ChangeDensity;
	str  << " ChangeVelocity " << sp.ChangeVelocity;
	str  << " del " << sp.deleted;
	str  << " ID  " << sp.ID;
	str  << " shapeID " << sp.shapeID;
	str  << " numNeighbors " << sp.numNeighbors;

	return str;
}	



//******************************************************************
// overloaded ofstream operator (file output)
//******************************************************************
ofstream & operator << ( ofstream & fStr, const SmoothParticle & sp)
{
	fStr << sp.pos;
	return fStr;
}	



//******************************************************************
// Calculates the pressure for this SPH particle.
// The state equation comes from **** and is used to model water.
//******************************************************************
void SmoothParticle::calcPressure(const Parameter & p)
{

	// this form is given in paper by Morris.
	pressure = p.c * p.c * density;

}



//******************************************************************
// Zeroes out the delta values. The reason is because the delta
// values come from summing up contributions from all neighbors and
// we don't want to add to the last time steps value.
//******************************************************************
void SmoothParticle::zeroSPHVars()
{
	ChangeDensity = 0.0;
	ChangeVelocity.setXY(0.0, 0.0);

	numNeighbors = 0;
}

//******************************************************************
// Adds the contributions of the change in density from the neighbor
// "sp".
//******************************************************************
void SmoothParticle::calcChangeDensity(const SmoothParticle & sp,
						Point v_ij)
{
	ChangeDensity += sp.getMass() *  
	(v_ij.getX() * gradientKernelX(sp) + 
	 v_ij.getY() * gradientKernelY(sp));
}	
		
	
//******************************************************************
// Add the contribution of the acceleration due to the free
// particle "sp".
//******************************************************************
void SmoothParticle::calcChangeVelocity(const SmoothParticle & sp, 
                                        Point v_ij, const Parameter & param)
{
	double pressTerm = pressureTerm(sp);
	double artVisc = artificialViscosity(sp, param);

	double oldChngVelX = ChangeVelocity.getX();
	double oldChngVelY = ChangeVelocity.getY();
		

	ChangeVelocity.setX(oldChngVelX + sp.getMass() *
						(pressTerm * gradientKernelX(sp) +
						v_ij.getX() * artVisc));


	ChangeVelocity.setY(oldChngVelY + sp.getMass() *
						(pressTerm * gradientKernelY(sp) +
						 v_ij.getY() * artVisc));

}
	

//******************************************************************
// Add the body force in the x direction.
//******************************************************************
void SmoothParticle::addForceX(const Parameter & param)
{
	ChangeVelocity.setX(ChangeVelocity.getX() + param.bodyFX);
}


//******************************************************************
// Add the body force in the y direction.
//******************************************************************
void SmoothParticle::addForceY(const Parameter & param)
{
	ChangeVelocity.setY(ChangeVelocity.getY() + param.bodyFY);
}


//******************************************************************
// Calculate the pressure gradient due to the free particle "sp".
//******************************************************************
double SmoothParticle::pressureTerm(const SmoothParticle & sp)
{
	return (-1.0 * (pressure/(density*density)
	    + sp.getPressure() / (sp.getDensity()*sp.getDensity())));
}



//******************************************************************
// This is used to simulate viscous fluids.
//******************************************************************
double SmoothParticle::artificialViscosity(const SmoothParticle & sp,
					 const Parameter & param)
{
	double distance = sqrt((pos.getX() - sp.getPositionX())*
						   (pos.getX() - sp.getPositionX()) +
						   (pos.getY() - sp.getPositionY())*
						   (pos.getY() - sp.getPositionY()));


	Point r_ij((pos.getX() - sp.getPositionX()), 
			   (pos.getY() - sp.getPositionY()));


	return (param.nu*(density + sp.getDensity()) *
		(r_ij.getX()*gradientKernelX(sp) + r_ij.getY()*gradientKernelY(sp))) /
		((density * sp.getDensity())* (distance*distance+0.01*h*h));

}


//******************************************************************
// Self explanatory.
//******************************************************************
double SmoothParticle::kernel(const SmoothParticle & sp)
{
	double distance;
	double normalization = 10.0 / (7.0*M_PI*h*h);	
	
	double dist1 = pos.getX() - sp.getPositionX();
	double dist2 = pos.getY() - sp.getPositionY();

	distance = sqrt(dist1 * dist1 + dist2 * dist2);

	if(distance < h)
		return (normalization) * (1.0 - 
		        1.5*distance*distance / (h*h) + 
				0.75*distance*distance*distance / (h*h*h));
	else if(distance < 2.0 * h)
		return  normalization * (2.0 - 3.0 * distance / h
		     + 1.5*distance*distance / (h*h) - 
			 0.25*distance*distance*distance / (h*h*h));
	else
		return 0.0;
		
}






//******************************************************************
// Self explanatory.
//******************************************************************
double SmoothParticle::gradientKernelX(const SmoothParticle & sp)
{
	double distance;
	double diffX, diffY;
	double normalization = 10.0 / (7.0*M_PI*h*h);	
	
	diffX = pos.getX() - sp.getPositionX();
	diffY = pos.getY() - sp.getPositionY();
	distance = sqrt(diffX*diffX + diffY*diffY);

	if(distance < h)
		return normalization * (-3.0/ (h*h) +
		 9.0*distance / (4.0*h*h*h)) * diffX;
	else if(distance < 2.0*h)
		return normalization * (-3.0/ (h*distance) +
		  3.0/(h*h) -3.0*distance/(4.0*h*h*h)) * diffX;
	else
		return 0.0;

		
}
	

//******************************************************************
// Self explanatory.
//******************************************************************
double SmoothParticle::gradientKernelY(const SmoothParticle & sp)
{
	double distance;
	double diffX, diffY;
	double normalization = 10.0 / (7.0*M_PI*h*h);	
	diffX = pos.getX() - sp.getPositionX();
	diffY = pos.getY() - sp.getPositionY();
	distance = sqrt(diffX*diffX + diffY*diffY);

	if(distance < h)
		return  normalization* (-3.0/ (h*h) +
		 9.0*distance / (4.0*h*h*h)) * diffY;
	else if(distance < 2.0*h)
		return  normalization* (-3.0/ (h*distance) +
		  3.0/(h*h) -3.0*distance/(4.0*h*h*h)) * diffY;
	else
		return 0.0;
		
}



//******************************************************************
// put all the necessary info about this particle into a packet 
// of doubles
//******************************************************************
void SmoothParticle::getPacket(double d[SMOOTH_SIZE])
{
	d[0] = pos.getX();
	d[1] = pos.getY();
	d[2] = vel.getX();
	d[3] = vel.getY();
	d[4] = mass;
	d[5] = getDensity();
	d[6] = getPressure();
	d[7] = getSmoothingLength();
	d[8] = getSpeedSound();
	d[9] = (double) deleted;
	d[10] = ChangeVelocity.getX();
 	d[11] = ChangeVelocity.getY();
	d[12] = ChangeDensity;
	d[13] = (double) ID;
	d[14] = (double) shapeID;
	d[15] = getMinDensity();
	d[16] = getMaxDensity();
	d[17] = (double) getNumNeighbors();
}

//******************************************************************
// put all the necessary info about this particle into a packet of floats
// used when we want to send the SP to the client
//******************************************************************
void SmoothParticle::getPacketClient(float d[SMOOTH_SIZE_CLIENT])
{
        d[0] = pos.getX();
        d[1] = pos.getY();
        d[2] = vel.getX();
        d[3] = vel.getY();
        d[4] = sqrt(vel.getX() * vel.getX() + vel.getY() * vel.getY());
        d[5] = getDensity();
        d[6] = getPressure();
        d[7] = (double) ID;
        d[8] = (double) shapeID;
}

//******************************************************************
// set this smoothparticle equal to the contents of our packet
//******************************************************************
void SmoothParticle::setPacket(double d[SMOOTH_SIZE])
{
	pos.setX(d[0]);
	pos.setY(d[1]);
	vel.setX(d[2]);
	vel.setY(d[3]);
	setMass(d[4]);
	setDensity(d[5]);
	setPressure(d[6]);
	setSmoothingLength(d[7]);
	setSpeedSound(d[8]);
	deleted = (int) d[9];

	Point pt(d[10], d[11]);
	setChangeVelocity(pt);

	setChangeDensity(d[12]);
	ID = (int) d[13];			
	shapeID = (int) d[14];			

	setMinDensity(d[15]);			
	setMaxDensity(d[16]);			

	setNumNeighbors((int) d[17]);
}

//******************************************************************
// Keeps track of our particle's min and max density
//******************************************************************
void SmoothParticle::setMinMaxDensity()
{
	if (density > maxDensity)
		maxDensity = density;
	else if (density < minDensity)
		minDensity = density;
}


//******************************************************************
// Used to keep a total of the number of neighbors within 2*h of
// this smooth particle.  Used for debugging.
//******************************************************************

void SmoothParticle::updateNumNeighbors(const SmoothParticle & sp)
{
	double distance = sqrt((pos.getX() - sp.getPositionX())*
						   (pos.getX() - sp.getPositionX()) +
						   (pos.getY() - sp.getPositionY())*
						   (pos.getY() - sp.getPositionY()));

	if (distance < 2*h)
		numNeighbors++;
}
