/* simple benchmark program to measure the 
   floating point performance on nodes in a 
   mpi environment

   Keith Vertanen 12/98 */

#include "server.h"

#include <stdio.h>
#include <mpi.h>

#include <stdlib.h>
#include <sys/times.h>
#include <pthread.h>

#define DEBUG 1
#define TAG 0
#define SIZE 100000     /* size of our benchmark arrays */
#define MAXHOSTS 32     /* max number of hosts we'll run on */


void start_timer(void);
double stop_timer(void);
long benchmark(void);
void run_benchmark(double result[], char hostname[MAXHOSTS][80], int
my_rank, int p);


main (int argc, char** argv)
{
	int my_rank;
	int p, i;
	double result[MAXHOSTS];
	char hostname[MAXHOSTS][80];
	int connected, command;
	int A[1], B[1];
	int end = 0;
	MPI_Status status;
	
	MPI_Init(&argc, &argv);
	MPI_Comm_rank(MPI_COMM_WORLD, &my_rank);
	MPI_Comm_size(MPI_COMM_WORLD, &p);

	if (my_rank==0)
	{
		/* masters code */

		/* setup a socket object */

		Server mylink(5009);
		connected = 0;

		while (!end)
		{
			if (!connected) 
			{
				/* wait for somebody to say hello */

				if (DEBUG)
					printf("Waiting for somebody...\n");

				mylink.connect();
				connected = 1;

				if (DEBUG)
					printf("Hello there!\n");
			}

			/* now we wait for the client to tell us what it wants */

			mylink.recv_ints(A, 1);
	
			if (A[0] == 0)
			{
				/* client wants to say bye bye */

				mylink.closesocket();
				connected = 0;
			}
			else if (A[0] == 1)
			{
				/* client wants a benchmark run */

				/* tell the slaves to run it too */

				if (DEBUG)
					printf("Telling everyone to run benchmark\n");

				for (i=1; i<p; i++)
					MPI_Send(A, 1, MPI_INT, i, TAG, MPI_COMM_WORLD);
					
				if (DEBUG)
					printf("Running benchmark\n");

				run_benchmark(result, hostname, my_rank, p);

				/* send the benchmark data to client */
				
				/* tell the client how many nodes running */
				B[0] = p;	
				mylink.send_ints(B, 1);

				/* send the hostnames */

				for (i=0; i<p; i++)
				{
					hostname[i][strlen(hostname[i])] = '\n';

					mylink.send_string(hostname[i]);
				}

				/* send the data */

				mylink.send_doubles(result, p);

			}
			else if (A[0] == 2)
			{
				/* client wants us to shutdown totally */

				mylink.closesocket();
				end = 1;

				/* tell the slaves to stop */
		
				for (i=1; i<p; i++)
					MPI_Send(A, 1, MPI_INT, i, TAG, MPI_COMM_WORLD);

			}			

		}
        }
	else
	{
		/* slave code */

		while (!end)
		{
			/* get our instructions from master */
			
			MPI_Recv(&command,1,MPI_INT,0,TAG,MPI_COMM_WORLD,&status);	

			if (command == 1)
			{
				/* time to run a benchmark */

				run_benchmark(result, hostname, my_rank, p);
			}

			else if (command == 2)
			{
				/* time to shutdown */
				
				end = 1;
			}		
		}
	}

	MPI_Finalize();
}

void run_benchmark(double result[], char hostname[MAXHOSTS][80], int
			my_rank, int p)
{
	int source;
        double totaltime, flops;
	long fp_ops = 0;

	MPI_Status status;

	/* everybody (host and slaves) measure the amount of time
	   it takes to do a bunch of floating point ops */

	start_timer();
	fp_ops += benchmark();
	totaltime = stop_timer();

	flops = (double) fp_ops / (double) totaltime / (double) 1000000;

	if (my_rank!=0)
	{
		/* slaves send result to master */

		/* get the name of our computer */
		char hostname[80];
		gethostname(hostname, 80);


		MPI_Send(&flops, 1, MPI_DOUBLE, 0, TAG, MPI_COMM_WORLD);
		MPI_Send(hostname, 80, MPI_CHAR, 0, TAG, MPI_COMM_WORLD);

	}
	else
	{
		/* master collects results from slaves */
		
		result[0] = (double) flops;
		
		gethostname(hostname[0], 80);

		if (DEBUG)
			printf("Gathering results...\n");

      		for (source=1; source<p; source++)
		{

			MPI_Recv(&result[source],1,MPI_DOUBLE,source,TAG,
				MPI_COMM_WORLD, &status);
			MPI_Recv(hostname[source],80,MPI_CHAR,source,TAG,
				MPI_COMM_WORLD, &status);

		}
	}
}


long benchmark()
{
	int i,j;

	int r, r2;
	double s, s2;

	/* fill our arrays with random values */

	for (i=0; i<SIZE; i++)
	{
		r = r + i / 3 * (i - 100);	/* 4 ops */
		s = s + (double) i / 3.343536 * (s - 12.12121212);   /* 4 ops */

		r2 = r*r*r*r*r / 4 / 5 / 6 / 7;                /* 8 ops */
		s2 = s*s*s*s*s / 4.0 / 5.0 / 1235.23 / 0.23123;   /* 8 ops */

		r2 = r2 + 12 + 34 + -1231 + 39823 - 2322 - 0 + 999 - 1; /* 8 ops */
		s2 = s2 + 12.0 + 34.45 - 438.345 + 23.0 - 633.33433434 +
					2.0 - 1.0 + 32.29829382982398;  /* 8 ops */
	}
	/* 40 * SIZE ops were done */

	return 40*SIZE;
}	


/*timer routines */

static struct tms	ignored;
static unsigned long time1;

void start_timer() {
	time1 = times(&ignored);
}

double stop_timer() {
	unsigned long time2;
	unsigned long totaltime;

	time2=times(&ignored);
	totaltime = time2-time1;
/*
	printf("Elasped time: %0.3f seconds\n",
		totaltime/(double)sysconf(_SC_CLK_TCK));
*/
	return totaltime/(double)sysconf(_SC_CLK_TCK);
}
